/* ful_util.c
 *   These have not been altered much since the original 
 *  version written by Crowell et al. I have streamlined the code a bit and 
 *  moved all the basic FUL functions to this file + "hidden" some functions
 *  that don't need to be global./KL
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ful_util.h"

#define MAX(x,y) ((x) > (y) ? x : y)
#define MIN(x,y) ((x) < (y) ? x : y)
#define SQUARED(x) ((x)*(x))

float old_ful(float *data, int size);
double narrow_chi_squared(float *data, int size, double k,
			  double max_diff, double est_fluct);
double modified_chi_squared(double flat_value, float *data, int size);

/*
 * modified_chi_squared (eq. 5)
 *
 * returns Sum_i((flat_value -data[i])^2)/Sum_i(flat_value)
 * where the first sum is carried out for data[i]<flat_value
 */

/* rewritten   DCR 10 Nov 2005
 */

double modified_chi_squared(double flat_value, float *data, int size)
{
  int   i;
  double mc = 0.0;

  for (i=0; i<size; i++) {
    if (data[i] < flat_value) mc += SQUARED(flat_value - data[i]);
  }
  return (mc / ((float) size * flat_value));
}

/*
 * old_ful  - WAS do_ful
 *
 * given yi = data and m = size,  return a value for k = FUL
 */

float old_ful(float *data, int size)
{
  int   i;
  float lowest, lo_ful, hi_ful, lo_mc, hi_mc, new_mc;
  float interp, x_range, max_allowed_err;

  lowest = data[0];
  for (i=1; i<size; i++) {  
    if (lowest > data[i]) lowest = data[i];
  }
  if (lowest < 0.0) fprintf(stderr, "ERROR in old_ful: negative data values...");
  lo_ful = lowest = MAX(0.001,lowest);
  hi_ful = lowest + sqrt((size + 4) * (lowest + 4.0));
  /* DCR: adding 4 to lowest in the above line fixes things for low statistics,
     i.e. lowest ~ 0, but doesn't screw up for large values of lowest.
     Adding 4 to size likewise helps deal with small data samples.
  */

  /* Interpolate to search for the correct value of ful.
     If we have n data-points and statistics/bin is ~ ful,
     then error bars on ful should be about sqrt(ful)/sqrt(n). We'll demand
     0.1 times this error,  or a relative error of 0.01,  whichever is lower.
  */

  lo_mc = modified_chi_squared(lo_ful, data, size);
  hi_mc = modified_chi_squared(hi_ful, data, size);
  interp = lo_ful;
  while (lo_mc < 0.5) {
    /* If linear interpolation yields something within 10% of the ends
       of the interval, don't trust it, because we might get stuck.
    */
    x_range = hi_ful - lo_ful;
    interp = lo_ful + (0.5 - lo_mc) * x_range / (hi_mc - lo_mc);
    interp = MIN(hi_ful - (0.1 *x_range), interp);
    interp = MAX(lo_ful + (0.1 *x_range), interp);

    max_allowed_err = 0.1 * sqrt(interp / size);
    max_allowed_err = MIN(0.01 * interp, max_allowed_err);
    if (x_range < max_allowed_err) break;

    new_mc = modified_chi_squared(interp, data, size);
    if (new_mc > 0.5) {
      hi_ful = interp;
      hi_mc = new_mc;
    } else {
      lo_ful = interp;
      lo_mc = new_mc;
    }
  }
  return interp;
}

float ful(float *data, int size, float fluct_fact)
{
  int    i, nn;
  double est_fluct, est_mean, y_max;
  float  ful;

  // return old_ful(data, size);

  ful = old_ful(data, size);
  est_mean  = MAX(0.5, ful);
  est_fluct = fluct_fact * sqrt(est_mean + 1.0);

  y_max = est_mean + 2.0 * est_fluct;
  /* calculate mean value below y_max */
  ful = 0.0;
  nn = 0;
  for (i = 0; i < size; i++) {
    if (data[i] <= y_max) {
      ful += data[i];
      nn++;
    }
  }
  ful /= (float) nn;

  est_mean  = ful;
  est_fluct = fluct_fact * sqrt(est_mean + 1.0);

  y_max = est_mean + est_fluct;
  /* calculate mean value below new y_max */
  ful = 0.0;
  nn = 0;
  for (i = 0; i < size; i++) {
    if (data[i] <= y_max) {
      ful += data[i];
      nn++;
    }
  }

  ful /= (float) nn;
  // ful += 0.35 * fluct_fact * sqrt(ful);
  return ful;
}


